----------------------------------
-- PACKAGES
----------------------------------

-- USED TO GROUP VARIABLES, PROCEDURES, FUNCTIONS, CURSORS, ETC
-- ADVANTAGES:
-- 1) MODULARITY AND ENCAPSULATION
-- 2) EASIER APPLICATION DESIGN
-- 3) BETTER PERFORMANCE: ORACLE PUT MULTIPLE OBJECTS IN MEMORY AT ONCE
-- 4) EASIER SECURITY ACCESS OF OBJECTS

-- PACKAGES ARE DIVIDED IN TWO PARTS:
-- A) SPECIFICATION (LIKE AN INTERFACE IN OOP)
-- B) BODY          (LIKE AN IMPLEMENTATION IN OOP)


-- A) PACKAGE SPECIFICATION
-- USED TO DECLARE PUBLIC VARIABLES AND OTHERS SUBPROGRAM'S SPECIFICATIONS
-- DOES NOT CONTAIN IMPLEMENTATION CODE
-- SEPARATES THE LOGIC FROM THE INTERFACE
-- CAN BE CREATED WITHOUT A PACKAGE BODY

-- SYNTAX:

CREATE [OR REPLACE] PACKAGE <name>
IS | AS
   <public types>
   <public variables>
   <subprograms specifications>
   ...

END;

-- EXAMPLE 1:

CREATE OR REPLACE PACKAGE PCK_1
IS
  g_PI CONSTANT NUMBER := 3.1415;
  
  PROCEDURE PROC_CALC (x NUMBER);
  FUNCTION  F_MATH    (x IN NUMBER, y IN NUMBER) RETURN INT;

END;

-- EXAMPLE 2: CREATING OVERLOADS


CREATE OR REPLACE PACKAGE PCK_2
AS
  g_PI NUMBER := 3.1415;
  
  PROCEDURE PROC_CALC (x IN NUMBER);
  PROCEDURE PROC_CALC (x IN NUMBER, Y OUT NUMBER);
  
  FUNCTION  F_MATH    (x NUMBER) RETURN INT;
  FUNCTION  F_MATH    (x NUMBER, y NUMBER) RETURN INT;

END;

-- DROP A PACKAGE SPECIFICATION

DROP PACKAGE  PCK_1;
DROP PACKAGE  PCK_2;

-- B) PACKAGE BODY
-- CONTAINS THE SUBPROGRAM'S DEFINITION
-- HOSTS THE LOGIC (IMPLEMENTATION) OF THE SUBPROGRAMS
-- CAN HAVE PRIVATE VARIABLES
-- REQUIRE A PACKAGE SPECIFICATION
-- INITIALIZE VARIABLES WITH A ONE-TIME PROCEDURE
-- REQUIRE FORWARDING DECLARION (ORDER OF SUBPROGRAMS)

-- SYNTAX:

CREATE [OR REPLACE] PACKAGE BODY <name>
IS | AS
   <private types>
   <private variables>
   <subprograms bodies>
   ...

END;

-- EXAMPLE 1: SIMPLE PACKAGE

CREATE OR REPLACE PACKAGE PCK_A
IS
  PROCEDURE PROC_CALC (x IN NUMBER);

END;

CREATE OR REPLACE PACKAGE BODY PCK_A
IS
   PROCEDURE PROC_CALC(X IN NUMBER)
   IS
   BEGIN
    DBMS_OUTPUT.PUT_LINE('x: ' || TO_CHAR(x*2) );

   END;
END;

-- EXECUTION
BEGIN
   PCK_A.PROC_CALC(10);
END;


-- EXAMPLE 2: OVERLOAD

CREATE OR REPLACE PACKAGE PCKB
IS
  PROCEDURE PROC_CALC (x IN NUMBER);
  PROCEDURE PROC_CALC (x IN NUMBER, y IN NUMBER);
  
END;

CREATE OR REPLACE PACKAGE BODY PCKB
IS
   PROCEDURE PROC_CALC(x IN NUMBER)
   IS
   BEGIN
    DBMS_OUTPUT.PUT_LINE('x: ' || TO_CHAR(x*2) );

   END;
   
   PROCEDURE PROC_CALC(x IN NUMBER, y IN NUMBER)
   IS
   BEGIN
    DBMS_OUTPUT.PUT_LINE('x*y: ' || TO_CHAR(x*y) );

   END;

END;


-- EXECUTION
BEGIN
   PCKB.PROC_CALC(3);
   PCKB.PROC_CALC(5,2);
END;


-- EXAMPLE 3: INITIALIZE VARIABLES WITH A ONE-TIME PROCEDURE

CREATE OR REPLACE PACKAGE PCKC
IS
  s_value NUMBER;

END;

CREATE OR REPLACE PACKAGE BODY PCKC
AS
BEGIN
   SELECT SALARY
   INTO s_value
   FROM HR.EMPLOYEES
   WHERE EMPLOYEE_ID = 176;
   
END;

-- EXECUTION
BEGIN
  DBMS_OUTPUT.PUT_LINE('Salary ' || TO_CHAR(PCKC.s_value) );
END;

-- DROP PACKAGE BODY

DROP PACKAGE BODY PCK_A;
DROP PACKAGE PCK_A;

DROP PACKAGE BODY PCKB;
DROP PACKAGE PCKB;

DROP PACKAGE BODY PCKC;
DROP PACKAGE PCKC;